# pylint: disable=relative-beyond-top-level
import unittest

from optimade.models import (
    StructureResponseMany,
    StructureResponseOne,
    ReferenceResource,
)

from ..utils import EndpointTestsMixin


class StructuresEndpointTests(EndpointTestsMixin, unittest.TestCase):

    request_str = "/structures"
    response_cls = StructureResponseMany

    def test_structures_endpoint_data(self):
        self.assertTrue("meta" in self.json_response)
        self.assertEqual(self.json_response["meta"]["data_available"], 17)
        self.assertEqual(self.json_response["meta"]["more_data_available"], False)
        self.assertTrue("data" in self.json_response)
        self.assertEqual(
            len(self.json_response["data"]),
            self.json_response["meta"]["data_available"],
        )

    def test_get_next_responses(self):
        total_data = self.json_response["meta"]["data_available"]
        page_limit = 5

        response = self.client.get(self.request_str + f"?page_limit={page_limit}")
        json_response = response.json()
        self.assertEqual(
            self.response.status_code,
            200,
            msg=f"Request failed: {self.response.json()}",
        )

        cursor = json_response["data"].copy()
        self.assertTrue(json_response["meta"]["more_data_available"])
        more_data_available = True
        next_request = json_response["links"]["next"]

        while more_data_available:
            next_response = self.client.get(next_request).json()
            next_request = next_response["links"]["next"]
            cursor.extend(next_response["data"])
            more_data_available = next_response["meta"]["more_data_available"]
            if more_data_available:
                self.assertEqual(len(next_response["data"]), page_limit)
            else:
                self.assertEqual(len(next_response["data"]), total_data % page_limit)

        self.assertEqual(len(cursor), total_data)


class SingleStructureEndpointTests(EndpointTestsMixin, unittest.TestCase):

    test_id = "mpf_1"
    request_str = f"/structures/{test_id}"
    response_cls = StructureResponseOne

    def test_structures_endpoint_data(self):
        self.assertTrue("data" in self.json_response)
        self.assertEqual(self.json_response["data"]["id"], self.test_id)
        self.assertEqual(self.json_response["data"]["type"], "structures")
        self.assertTrue("attributes" in self.json_response["data"])
        self.assertTrue("_exmpl_chemsys" in self.json_response["data"]["attributes"])


class MissingSingleStructureEndpointTests(EndpointTestsMixin, unittest.TestCase):

    test_id = "mpf_random_string_that_is_not_in_test_data"
    request_str = f"/structures/{test_id}"
    response_cls = StructureResponseOne

    def test_structures_endpoint_data(self):
        self.assertTrue("data" in self.json_response)
        self.assertTrue("meta" in self.json_response)
        self.assertEqual(self.json_response["data"], None)
        self.assertEqual(self.json_response["meta"]["data_returned"], 0)
        self.assertEqual(self.json_response["meta"]["more_data_available"], False)


class SingleStructureWithRelationshipsTests(EndpointTestsMixin, unittest.TestCase):

    test_id = "mpf_1"
    request_str = f"/structures/{test_id}"
    response_cls = StructureResponseOne

    def test_structures_endpoint_data(self):
        self.assertTrue("data" in self.json_response)
        self.assertEqual(self.json_response["data"]["id"], self.test_id)
        self.assertEqual(self.json_response["data"]["type"], "structures")
        self.assertTrue("attributes" in self.json_response["data"])
        self.assertTrue("relationships" in self.json_response["data"])
        self.assertDictEqual(
            self.json_response["data"]["relationships"],
            {"references": {"data": [{"type": "references", "id": "dijkstra1968"}]}},
        )
        self.assertTrue("included" in self.json_response)
        self.assertEqual(
            len(self.json_response["data"]["relationships"]["references"]["data"]),
            len(self.json_response["included"]),
        )

        ReferenceResource(**self.json_response["included"][0])


class MultiStructureWithSharedRelationshipsTests(EndpointTestsMixin, unittest.TestCase):

    request_str = "/structures?filter=id=mpf_1 OR id=mpf_2"
    response_cls = StructureResponseMany

    def test_structures_endpoint_data(self):
        # mpf_1 and mpf_2 both contain the same reference relationship, so response should not duplicate it
        self.assertTrue("data" in self.json_response)
        self.assertEqual(len(self.json_response["data"]), 2)
        self.assertTrue("included" in self.json_response)
        self.assertEqual(len(self.json_response["included"]), 1)


class MultiStructureWithRelationshipsTests(EndpointTestsMixin, unittest.TestCase):

    request_str = "/structures?filter=id=mpf_1 OR id=mpf_23"
    response_cls = StructureResponseMany

    def test_structures_endpoint_data(self):
        # mpf_4 contains no relationships, which shouldn't break anything
        self.assertTrue("data" in self.json_response)
        self.assertEqual(len(self.json_response["data"]), 2)
        self.assertTrue("included" in self.json_response)
        self.assertEqual(len(self.json_response["included"]), 1)


class MultiStructureWithOverlappingRelationshipsTests(
    EndpointTestsMixin, unittest.TestCase
):

    request_str = "/structures?filter=id=mpf_1 OR id=mpf_3"
    response_cls = StructureResponseMany

    def test_structures_endpoint_data(self):
        self.assertTrue("data" in self.json_response)
        self.assertEqual(len(self.json_response["data"]), 2)
        self.assertTrue("included" in self.json_response)
        self.assertEqual(len(self.json_response["included"]), 2)


class SingleStructureEndpointEmptyTest(EndpointTestsMixin, unittest.TestCase):

    test_id = "non_existent_id"
    request_str = f"/structures/{test_id}"
    response_cls = StructureResponseOne

    def test_structures_endpoint_data(self):
        self.assertTrue("data" in self.json_response)
        self.assertEqual(self.json_response["data"], None)



# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from LAMMPS.BaseClasses import Region


class Block(Region):
    name = 'block'

    def __init__(self, *args):
        pass


class Cone(Region):
    name = 'cone'

    def __init__(self, *args):
        pass


class Cylinder(Region):
    name = 'cylinder'

    def __init__(self, *args):
        pass


class Plane(Region):
    name = 'plane'

    def __init__(self, *args):
        pass


class Prism(Region):
    name = 'prism'

    def __init__(self, *args):
        pass


class Sphere(Region):
    name = 'sphere'

    def __init__(self, *args):
        pass


class Union(Region):
    name = 'union'

    def __init__(self, *args):
        pass


class Intersect(Region):
    name = 'intersect'

    def __init__(self, *args):
        pass


if __name__ == '__main__':
    pass



# Copyright 2016-2018 Arvid Conrad Ihrig, Aliaksei Mazheika
#                     Fritz-Haber-Institut der Max-Planck-Gesellschaft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import numpy as np
from nomadcore.simple_parser import SimpleMatcher as SM
import nomadcore.elements as elements
from turbomoleparser.TurbomoleCommon import RE_FLOAT

logger = logging.getLogger("nomad.turbomoleParser")


class _Atom(object):

    def __init__(self, x, y, z, elem, charge, shells, isotope, pseudo):
        self.x = float(x)
        self.y = float(y)
        self.z = float(z)
        self.elem = elem.capitalize()
        self.charge = float(charge)
        self.shells = int(shells) if shells else -1
        self.isotope = int(isotope)
        if pseudo:
            self.is_pseudo = True if pseudo == "1" else False
        else:
            self.is_pseudo = None  # effective core potential not specified
        self.label = self.elem if self.shells > 0 else self.elem+"_1"


class _AtomKind(object):

    def __init__(self, index, elem, charge, isotope, pseudo):
        self.index = int(index)
        self.elem = elem.capitalize()
        self.charge = float(charge)
        self.isotope = int(isotope)
        if pseudo:
            self.is_pseudo = True if pseudo == "1" else False
        else:
            self.is_pseudo = None  # effective core potential not specified


class BasisSet(object):

    def __init__(self, name, index, cartesian, num_atoms):
        self.name = name
        self.index = index
        self.cartesian = cartesian
        self.num_atoms = int(num_atoms)


class SystemParser(object):

    def __init__(self, context, key="geo"):
        context[key] = self
        self.__context = context
        self.__backend = None
        self.__index_basis_set = -1
        self.__atoms = list()
        self.__atom_kinds = dict()
        self.__basis_sets = dict()
        self.__basis_one_to_one = True
        self.__auxbasis_sets = dict()
        self.__auxbasis_one_to_one = True

    def purge_data(self):
        self.__index_basis_set = -1
        self.__atoms = list()
        self.__atom_kinds = dict()
        self.__basis_sets = dict()
        self.__basis_one_to_one = True
        self.__auxbasis_sets = dict()
        self.__auxbasis_one_to_one = True

    def set_backend(self, backend):
        self.__backend = backend

    # match builders

    def num_atoms(self):
        return len(self.__atoms)

    def write_atomic_data(self, system_index):
        pos = np.ndarray(shape=(len(self.__atoms), 3), dtype=float)
        labels = list()
        atom_numbers = np.ndarray(shape=(len(self.__atoms),), dtype=float)
        self.__backend.addValue("number_of_atoms", len(self.__atoms), system_index)
        for i, atom in enumerate(self.__atoms):
            pos[i, 0:3] = (atom.x, atom.y, atom.z)
            labels.append(atom.elem)
            atom_numbers[i] = elements.get_atom_number(atom.elem)
        self.__backend.addArrayValues("atom_positions", pos, gIndex=system_index, unit="bohr")
        self.__backend.addArrayValues("atom_labels", np.asarray(labels, dtype=str),
                                      gIndex=system_index)
        self.__backend.addArrayValues("atom_atom_number", atom_numbers, gIndex=system_index)

    def write_basis_set_mapping(self, config_index, method_index):
        # TODO: if these conditions are not true, we need to read the files basis and coords instead
        if self.__basis_one_to_one and len(self.__basis_sets) > 0:
            self.__write_simple_method_basis_set_mapping(method_index, False, self.__basis_sets)
        if self.__auxbasis_one_to_one and len(self.__auxbasis_sets) > 0:
            self.__write_simple_method_basis_set_mapping(method_index, True, self.__auxbasis_sets)
        if sum(x.num_atoms for x in self.__basis_sets.values()) == len(self.__atoms):
            self.__write_simple_atomic_basis_set(config_index, False, self.__basis_sets)
        if sum(x.num_atoms for x in self.__auxbasis_sets.values()) == len(self.__atoms):
            self.__write_simple_atomic_basis_set(config_index, True, self.__auxbasis_sets)

    def __write_simple_atomic_basis_set(self, config_index, is_auxiliary, basis_sets):
        references = {"section_single_configuration_calculation": config_index}
        index = self.__backend.openSection("section_basis_set")
        self.__backend.setSectionInfo("section_basis_set", index, references)
        mapping = np.ndarray(shape=(len(self.__atoms),), dtype=int)
        for i, atom in enumerate(self.__atoms):
            mapping[i] = basis_sets[atom.elem].index
        self.__backend.addArrayValues("mapping_section_basis_set_atom_centered", mapping)
        self.__backend.addValue("basis_set_kind", "density" if is_auxiliary else "wavefunction",
                                index)
        self.__backend.closeSection("section_basis_set", index)

    def __write_simple_method_basis_set_mapping(self, method_index, is_auxiliary, basis_sets):
        references = {"section_method": method_index}
        index = self.__backend.openSection("section_method_basis_set")
        self.__backend.setSectionInfo("section_method_basis_set", index, references)
        self.__backend.addValue("number_of_basis_sets_atom_centered", len(self.__atom_kinds), index)
        refs = np.ndarray(shape=(len(self.__atom_kinds),2), dtype=int)
        refs[:, :] = -1
        for i, elem in enumerate(sorted(self.__atom_kinds)):
            kind = self.__atom_kinds[elem]
            refs[i, 0] = kind.index
            refs[i, 1] = basis_sets[kind.elem].index
        self.__backend.addValue("method_basis_set_kind",
                                "density" if is_auxiliary else "wavefunction", index)
        self.__backend.addArrayValues("mapping_section_method_basis_set_atom_centered", refs, index)
        self.__backend.closeSection("section_method_basis_set", index)

    def build_qm_geometry_matcher(self):

        def add_atom(backend, groups):
            new_atom = _Atom(x=groups[0], y=groups[1], z=groups[2], elem=groups[3],charge=groups[5],
                             isotope=groups[7], shells=groups[4], pseudo=groups[6])
            self.__atoms.append(new_atom)
            if groups[3].capitalize() not in self.__atom_kinds:
                index = backend.openSection("section_method_atom_kind")
                kind = _AtomKind(index=index, elem=groups[3],charge=groups[5],
                                 isotope=groups[7], pseudo=groups[6])
                atom_number = elements.get_atom_number(kind.elem)
                backend.addValue("method_atom_kind_atom_number", atom_number, index)
                # backend.addValue("method_atom_kind_explicit_electrons", ..., index)
                backend.addValue("method_atom_kind_label", kind.elem, index)
                # backend.addValue("method_atom_kind_mass", ..., index)
                # backend.addValue("method_atom_kind_pseudopotential_name", ..., index)
                backend.closeSection("section_method_atom_kind", index)
                self.__atom_kinds[kind.elem] = kind

        # x, y, z, element, (shells), charge, (pseudo), isotope
        atom_re = r"\s*" + 3 * ("("+RE_FLOAT+")\s+") + r"([a-zA-Z]+)(\s+[0-9]+)?" \
                  r"\s+("+RE_FLOAT+")(\s+[-+]?[0-9]+)?\s+([-+]?[0-9]+)"
        atom = SM(atom_re, repeats=True, name="single atom", startReAction=add_atom)
        header_re = r"\s*atomic\s+coordinates\s+atom(?:\s+shells)?\s+charge(?:\s+pseudo)?\s+isotop"

        return SM(name="geometry",
                  startReStr=r"\s*\|\s+Atomic coordinate, charge and isotope? information\s+\|",
                  subMatchers=[
                      SM(r"\s*-{20}-*", name="<format>", coverageIgnore=True),
                      SM(header_re, name="atom list", subMatchers=[atom])
                  ]
                  )

    def build_qm_geometry_matcher_statpt(self):

        def add_atom(backend, groups):
            new_atom = _Atom(x=groups[1], y=groups[2], z=groups[3], elem=groups[0], charge=0,
                             isotope=0, shells=None, pseudo=None)
            self.__atoms.append(new_atom)

        atom = SM(r"\s*[0-9]+\s+([A-z]+)"+ 3 * (r"\s+("+RE_FLOAT+")") + "\s*$",
                  name="single atom",
                  repeats=True,
                  startReAction=add_atom)

        return SM(name="geometry",
                  startReStr=r"\s*ATOM\s+CARTESIAN\s+COORDINATES\s*$",
                  subMatchers=[
                      atom
                  ]
                  )

    def build_orbital_basis_matcher(self, title_regex=None):
        if not title_regex:
            title_regex = r"\s*\|\s*basis\s+set\s+information\s*\|"
        return self.__build_basis_matcher(False, title_regex)

    def build_auxiliary_basis_matcher(self, title_regex=None):
        if not title_regex:
            title_regex = r"\s*\|\s*Auxiliary\s+basis\s+set\s+information\s*\|"
        return self.__build_basis_matcher(True, title_regex)

    def __build_basis_matcher(self, is_auxbasis, title_regex):
        basis_type = "Auxiliary Basis" if is_auxbasis else "Orbital Basis"

        class LocalBasisData(object):
            spherical = False

        def set_spherical_basis(backend, groups):
            LocalBasisData.spherical = True

        def add_basis_set(backend, groups):
            # TODO: support assignment of different basis sets to atoms of the same element
            index = backend.openSection("section_basis_set_atom_centered")
            key = groups[0].capitalize()
            basis_set = BasisSet(name=groups[4], index=index,
                                 cartesian=not LocalBasisData.spherical, num_atoms=groups[1])
            if is_auxbasis:
                if key in self.__auxbasis_sets:
                    self.__auxbasis_one_to_one = False
                else:
                    self.__auxbasis_sets[key] = basis_set
            else:
                if key in self.__basis_sets:
                    self.__basis_one_to_one = False
                else:
                    self.__basis_sets[key] = basis_set
            atom_number = elements.get_atom_number(groups[0].capitalize())
            backend.addValue("basis_set_atom_centered_short_name", groups[4], index)
            backend.addValue("basis_set_atom_number", atom_number, index)
            if LocalBasisData.spherical:
                backend.addValue("number_of_basis_functions_in_basis_set_atom_centered",
                                 int(groups[3]), index)
            else:
                logger.warning("no basis function count information for cartesian Gaussians!")
            backend.closeSection("section_basis_set_atom_centered", index)

        # elem, num atoms, prim gauss, contracted gauss, name, contracted details, prim details
        basis_re = r"\s*([A-z]+)\s+([0-9]+)\s+" \
                   r"([0-9]+)\s+([0-9]+)\s+" \
                   r"([A-z0-9-\(\)]+)\s+" \
                   r"\[((?:[0-9]+[spdfghij])+)\|((?:[0-9]+[spdfghij])+)\]"
        basis = SM(basis_re, repeats=True, name="basis assignment", startReAction=add_basis_set)
        basis_regex = r"1s\s+" + \
            "".join((r"(?:"+str(2*(i+1)+1)+x+r"\s+)?" for i,x in enumerate("pdfghijkl")))
        basis_type = "(\s+auxiliary)?\s+basis\s+set"
        header_re = r"\s*we\s+will\s+work\s+with\s+the\s+"+basis_regex+"\s*\.\.\." + basis_type
        gauss_type_spherical = SM(header_re,
                                  name="spherical Gaussians",
                                  subMatchers=[
                                      SM(r"\s*...i.e. with spherical basis functions...",
                                         name="spherical Gaussians",
                                         required=True)
                                  ],
                                  startReAction=set_spherical_basis
                                  )

        return SM(title_regex,
                  name=basis_type,
                  subMatchers=[
                      SM(r"\s*\+----*\+", name="<format>", coverageIgnore=True),
                      gauss_type_spherical,
                      SM(r"\s*type\s+atoms\s+prim\s+cont\s+basis",
                         name=basis_type),
                      SM(r"\s*----*", name="<format>", coverageIgnore=True),
                      basis,
                      SM(r"\s*----*", name="<format>", coverageIgnore=True),
                      SM(r"\s*total:\s*([0-9]+)\s+([0-9]+)\s+([0-9]+)",
                         name=basis_type),
                      SM(r"\s*----*", name="<format>", coverageIgnore=True),
                      SM(r"\s*total number of primitive shells\s*:\s*([0-9]+)",
                         name=basis_type),
                      SM(r"\s*total number of contracted shells\s*:\s*([0-9]+)",
                         name=basis_type),
                      SM(r"\s*total number of cartesian basis functions\s*:\s*([0-9]+)",
                         name=basis_type),
                      SM(r"\s*total number of SCF-basis functions\s*:\s*([0-9]+)",
                         name=basis_type),
                  ]
                  )



# Copyright 2019-2018 Markus Scheidgen
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import sys
import numpy as np

from ase.io import read as ase_read

from nomadcore.simple_parser import SimpleMatcher
from nomadcore.baseclasses import ParserInterface, MainHierarchicalParser

from nomad.parsing import Backend


"""
A very basic VASP OUTCAR parser. It is only used to get nomad repository
metadata in absence of the VASP .xml file.
"""


class VaspOutcarParser(ParserInterface):

    def get_metainfo_filename(self):
        return 'vasp.nomadmetainfo.json'

    def get_parser_info(self):
        return {
            'name': 'vaspoutcar_parser',
            'version': '1.0.0'
        }

    def setup_version(self):
        self.setup_main_parser(None)

    def setup_main_parser(self, _):
        self.main_parser = MainParser(self.parser_context)


class MainParser(MainHierarchicalParser):
    def __init__(self, parser_context, *args, **kwargs):
        super().__init__(parser_context, *args, **kwargs)
        self.lattice_vectors = []

        self.root_matcher = SimpleMatcher(
            name='root',
            startReStr=r'vasp.',
            weak=True,
            sections=['section_run', 'section_system', 'section_method'],
            subMatchers=[
                SimpleMatcher(r'\svasp.(?P<program_version>\d+.\d+.\d+)\s.*')
            ]
        )

    def parse(self, mainfile, *args, **kwargs):
        self.ase = ase_read(mainfile, format='vasp-out')
        super().parse(mainfile, *args, **kwargs)

    def add_lattice_vector(self, backend, data):
        self.lattice_vectors.append(list(float(x.strip()) for x in data[1].split(',')))

    def onClose_section_method(self, backend, *args, **kwargs):
        backend.openNonOverlappingSection('section_XC_functionals')
        backend.addValue('XC_functional_name', 'GGA_X_PBE')
        backend.closeNonOverlappingSection('section_XC_functionals')
        backend.addValue('electronic_structure_method', 'DFT')

    def onClose_section_system(self, backend, *args, **kwargs):
        backend.addArrayValues('atom_labels', np.array(self.ase.get_chemical_symbols()))
        backend.addArrayValues('atom_positions', self.ase.get_positions() * 1e-10)
        backend.addArrayValues('lattice_vectors', self.ase.get_cell() * 1e-10)
        backend.addArrayValues('configuration_periodic_dimensions', self.ase.get_pbc())

    def onClose_section_run(self, backend, *args, **kwargs):
        backend.addValue('program_name', 'VASP')
        backend.addValue('program_basis_set_type', 'plane waves')


if __name__ == "__main__":
    parser = VaspOutcarParser(backend=Backend)
    parser.parse(sys.argv[1])
    print(parser.parser_context.super_backend)



#!/usr/bin/env python
# Copyright 2016-2017 Henning Glawe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nomadcore.simple_parser import mainFunction, SimpleMatcher as SM, CachingLevel
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
import os
import sys
import json
import re
import logging
import nomadcore.unit_conversion.unit_conversion as unit_conversion
import math
import numpy as np
from . import FploCommon as FploC
import calendar
from nomadcore.parser_backend import valueForStrValue
from .FploCommon import RE_f, RE_i, cRE_f, cRE_i
from nomadcore.parser_backend import valueForStrValue
from . import FploInputParser


LOGGER = logging.getLogger(__name__)


# lookup table translating relativistic treatment from fplo to NOMAD
FPLO_RELATIVISTIC = {
    'non': '',
    'scalar': 'scalar_relativistic',
    'KH scalar': 'scalar_relativistic',
    'full': '4_component_relativistic',
}

# lookup table translating XC functional from fplo to NOMAD
FPLO_XC_FUNCTIONAL = {
    'Exchange only                (LSDA)': [
        { 'XC_functional_name': 'LDA_X' },
    ],
    'von Barth Hedin              (LSDA)': [
        { 'XC_functional_name': 'LDA_X' },
        { 'XC_functional_name': 'LDA_C_VBH' },
    ],
    'Perdew and Zunger            (LSDA)': [
        { 'XC_functional_name': 'LDA_X' },
        { 'XC_functional_name': 'LDA_C_PZ' },
    ],
    'Perdew Wang 92               (LSDA)': [
        { 'XC_functional_name': 'LDA_X' },
        { 'XC_functional_name': 'LDA_C_PW' },
    ],
    'Perdew Burke Ernzerhof 96    (GGA)':  [
        { 'XC_functional_name': 'GGA_X_PBE' },
        { 'XC_functional_name': 'GGA_C_PBE' },
    ],
}

FPLO_DFT_PLUS_U_PROJECTION_TYPE = {
    'orthogonal': 'orthogonalized atomic',
}

FPLO_DFT_PLUS_U_FUNCTIONAL = {
    'LSDA+U Atomic limit      (AL)': 'fully localized limit',
    'LSDA+U Around mean field (AMF/OP)': 'around mean field',
}

class ParserFplo14(object):
    """main place to keep the parser status, open ancillary files,..."""
    def __init__(self, metaInfoEnv):
        self.parserInfo = FploC.PARSER_INFO_DEFAULT.copy()
        self.cachingLevelForMetaName = {}
        for name in metaInfoEnv.infoKinds:
            # set all temporaries to caching-only
            if name.startswith('x_fplo_t_'):
                self.cachingLevelForMetaName[name] = CachingLevel.Cache
        self.coverageIgnoreList = [
            # ignore empty lines
            r"\s*",
            # table separators
            r"^\s*[=%-]+\s*$",
            r"^\s*\|\s*\|\s*$",
            # table separators in LSDA+U
            r"^\s*LSDA\+U:\s*[=%-]+\s*$",
        ]
        self.coverageIgnore = None
        unit_conversion.register_userdefined_quantity('usrTpibbohr', '1/bohr', 2*math.pi)

    def parse(self):
        self.coverageIgnore = re.compile(r"^(?:" + r"|".join(self.coverageIgnoreList) + r")$")
        mainFunction(self.mainFileDescription(), metaInfoEnv, self.parserInfo,
                     cachingLevelForMetaName=self.cachingLevelForMetaName,
                     superContext=self)

    def mainFileDescription(self):
        # assemble matchers and submatchers
        result = SM(name='root',
            startReStr=r"$",
            subMatchers=[
                SM(name='newrun', repeats=True, forwardMatch=True,
                   startReStr=r"\s*\|\s*FULL-POTENTIAL LOCAL-ORBITAL MINIMUM BASIS BANDSTRUCTURE CODE\s*\|\s*$",
                   sections=['section_run'],
                   fixedStartValues={
                       'program_name': 'fplo',
                       'program_basis_set_type': 'local-orbital minimum-basis',
                   },
                   subMatchers=[
                       SM(name='method_structure',
                          # parent has forwardMatch set, allows to
                          # replicate for opening method/system sections, both closed by SCF matcher
                          startReStr=r"\s*\|\s*FULL-POTENTIAL LOCAL-ORBITAL MINIMUM BASIS BANDSTRUCTURE CODE\s*\|\s*$",
                          sections=['section_method', 'section_system'],
                          subMatchers=[
                          ] + self.SMs_header() + [
                          ] + self.SMs_input() + [
                          ] + self.SMs_crystal_structure() + [
                          ] + self.SMs_method() + [
                          ],
                       ),
                       SM(name='start_scf',
                          startReStr=r"\s*SCF: Iteration version\s*\(\s*\d+\) : Iterat\s*:\s*(?:.*?)\s*$",
                          sections=['section_single_configuration_calculation'],
                          subMatchers=self.SMs_scf(),
                       ),
                   ]
                ),
            ]
        )
        return result

    def SMs_header(self):
        result = [
            SM(name='copyrightspam', repeats=True, coverageIgnore=True,
               startReStr=r"\s*\|\s*(?:" + r"|".join([
                   r"FULL RELATIVISTIC VERSION",
                   r"by K\. Koepernik, A\.Ernst and H\.Eschrig \(2003\)",
                   r"relativistic version by Ingo Opahle",
                   r"LSDA\+U by Igor Chaplygin",
                   r"Any publication of results obtained by this program",
                   r"has to include the citation:",
                   r"K\.Koepernik and H\.Eschrig, Phys\. Rev\. B 59, 1743 \(1999\)",
                   r"Any publication of CPA results obtained by this program",
                   r"additionally has to include the citation:",
                   r"K\. Koepernik, B\. Velicky, R\. Hayn and H\. Eschrig,",
                   r"Phys\. Rev\. B 55, 5717 \(1997\)",
               ]) + r")\s*\|\s*$",
            ),
            SM(name='versMain',
               startReStr=r"\s*\|\s*main\s+version\s*:\s*(?P<x_fplo_t_program_version_main>\S+)\s*\|\s*$",
               subMatchers=[
                   SM(name='versSub',
                      startReStr=r"\s*\|\s*sub\s+version\s*:\s*(?P<x_fplo_program_version_sub>\S+)\s*\|\s*$",
                   ),
                   SM(name='versRelease',
                      startReStr=r"\s*\|\s*release\s*:\s*(?P<x_fplo_t_program_version_release>\S+)\s*\|\s*$",
                   ),
                   SM(name='compileOpts', repeats=True,
                      startReStr=r"\s*\|\s*compiled with\s*(?P<x_fplo_program_compilation_options>.*?)\s*\|\s*$",
                   ),
                   SM(name='runDate',
                      startReStr=r"\s*\|\s*date\s*:\s*(?P<time_run_date_start__strFploDate>.+?)\s*\|\s*$",
                   ),
                   SM(name='runHost',
                      startReStr=r"\s*\|\s*host\s*:\s*(?P<x_fplo_t_run_hosts>.+?)\s*\|\s*$"
                   ),
               ],
            ),
        ]
        return result

    def SMs_sym_msg(self):
        # msg appears twice, so new generator
        result = [
            SM(name='inpSymInfo',
               startReStr=r"\s*INFORMATION in MODULE SYMMETRY\(crystal_structure_setup\):\s*$",
               subMatchers=[
                   SM(name='InfoHexAxis1',
                      startReStr=r"\s*INFORMATION: (?P<message_info_run>\(makeunitcell\): Third hexagonal axis angle != 120 degree!)\s*$",
                   ),
                   SM(name='InfoHexAxis2',
                      startReStr=r"\s*(?P<message_info_run>I will correct it !)\s*$",
                   ),
               ],
            ),
        ]
        return result

    def adHoc_input_content(self, parser):
        LOGGER.error("TODO: parse C-like echoed input")
        input_parser = FploInputParser.FploInputParser(
            None,
            annotated_line_callback=self.callback_annotated_input_line)
        while True:
            fInLine = parser.fIn.readline()
            if re.match(r'^\s*-{60,}\s*$', fInLine):
                self.callback_annotated_input_line(fInLine)
                break
            input_parser.parse_line(fInLine)
        input_parser.onEnd_of_file()

    def callback_annotated_input_line(self, annotated_input_line):
        minfo = {
            # raw line
            'fInLine': '', # fInLine,
            'fInLineNr': self.parser.fIn.lineNr,
            # information about SimpleMatcher
            'matcherName': 'tokenizer',
            'defFile': 'FploInputParser.py',
            'defLine': 0,
            'matcher_does_nothing': False,
            'which_re': 'tokenizer',
            # classification of match
            'matchFlags': 1,
            'match': 3, # 0 - no, 1 - partial, 3 - full
            'coverageIgnore': 0, # 0 - no, 1 - local, 3 - global
            # overall span of match, and spans of group captures
            'span': [],
            # capture group names
            'matcherGroup': [],
            # we have pre-highlighted line
            'highlighted': annotated_input_line,
        }
        self.parser.annotator.annotate(minfo)

    def SMs_input(self):
        result = [
            SM(name='inpSym',
               startReStr=r"\s*File =\.sym exists!\s*$",
               subMatchers=[
               ] + self.SMs_sym_msg() + [
               ],
            ),
            SM(name='inpIn',
               startReStr=r"\s*File =\.in exists!\s*$",
               subMatchers=[
                   SM(name='inpInCompound',
                      startReStr=r"\s*Compound\s*:\s*(?P<system_name>.*?)\s*$",
                   ),
                   SM(name='startEchoInput',
                      startReStr=r"\s*Start: content of =.in\s*$",
                      subMatchers=[
                          SM(name='startEchoInputDelimiter',
                             startReStr=r'^\s*-{60,}\s*$',
                             adHoc=self.adHoc_input_content,
                          ),
                      ],
                   ),
               ] + self.SMs_sym_msg() + [
               ],
            ),
        ]
        return result

    def adHoc_cs_structure_type(self, parser):
        if parser.lastMatch['x_fplo_structure_type'] == 'Crystal':
            parser.backend.addArrayValues('configuration_periodic_dimensions',
                                          np.array([True, True, True]))
        elif parser.lastMatch['x_fplo_structure_type'] == 'Molecule':
            parser.backend.addArrayValues('configuration_periodic_dimensions',
                                          np.array([False, False, False]))
        else:
            raise RuntimeError('unexpected value for x_fplo_structure_type: ',
                               parser.lastMatch['x_fplo_structure_type'])

    def SMs_crystal_structure(self):
        result = [
            SM(name='csHead',
               startReStr=r"\s*CALCULATION OF CRYSTALL STRUCTURE\s*$",
               subMatchers=[
                   SM(name='csInputHead',
                      startReStr=r"\s*INPUT DATA\s*$",
                   ),
                   SM(name='csSymmHead',
                      startReStr=r"\s*SYMMETRY CREATION\s*$",
                      subMatchers=[
                          SM(name='csLatticeVectors',
                             startReStr=r"\s*lattice vectors\s*$",
                             subMatchers=[
                                 SM(name='csLatA1',
                                    startReStr=r"\s*a1\s*:\s+" + FploC.re_vec('x_fplo_t_vec_a', 'bohr') + r"\s*$",
                                 ),
                                 SM(name='csLatA2',
                                    startReStr=r"\s*a2\s*:\s+" + FploC.re_vec('x_fplo_t_vec_a', 'bohr') + r"\s*$",
                                 ),
                                 SM(name='csLatA3',
                                    startReStr=r"\s*a3\s*:\s+" + FploC.re_vec('x_fplo_t_vec_a', 'bohr') + r"\s*$",
                                 ),
                             ],
                          ),
                          SM(name='csReciprocalVectors',
                             startReStr=r"\s*reciprocial lattice vectors / 2\*Pi\s*$",
                             subMatchers=[
                                 SM(name='csLatB1',
                                    startReStr=r"\s*g1\s*:\s+" + FploC.re_vec('x_fplo_t_vec_b', 'usrTpibbohr') + r"\s*$",
                                 ),
                                 SM(name='csLatB2',
                                    startReStr=r"\s*g2\s*:\s+" + FploC.re_vec('x_fplo_t_vec_b', 'usrTpibbohr') + r"\s*$",
                                 ),
                                 SM(name='csLatB3',
                                    startReStr=r"\s*g3\s*:\s+" + FploC.re_vec('x_fplo_t_vec_b', 'usrTpibbohr') + r"\s*$",
                                 ),
                             ],
                          ),
                      ],
                   ),
                   SM(name='csAtomsHead',
                      startReStr=r"\s*UNIT CELL CREATION\s*$",
                      subMatchers=[
                          SM(name='csAtomsPositionHead',
                             startReStr=r"\s*No.  Element WPS CPA-Block\s+X\s+Y\s+Z\s*$",
                             subMatchers=[
                                 SM(name='csAtomPositions', repeats=True,
                                    startReStr=(
                                        r"\s*(?P<x_fplo_t_atom_idx>" + RE_i + r")" +
                                        r"\s+(?P<x_fplo_t_atom_labels>\S+)" +
                                        r"\s+(?P<x_fplo_t_atom_wyckoff_idx>" + RE_i + r")" +
                                        r"\s+(?P<x_fplo_t_atom_cpa_block>" + RE_i + r")" +
                                        r"\s+" + FploC.re_vec('x_fplo_t_atom_positions', 'bohr') +
                                        r"\s*$"
                                    ),
                                 ),
                             ],
                          ),
                      ],
                   ),
                   SM(name='csStructureType',
                      startReStr=r"Structure type:\s*(?P<x_fplo_structure_type>\S+)\s*$",
                      adHoc=self.adHoc_cs_structure_type
                   ),
               ],
            ),
        ]
        return result

    def SMs_method(self):
        result = [
            SM(name='mRelativistic',
               startReStr=r"\s*(?P<x_fplo_t_relativity_method>.*?)\s*relativistic\s*calculation!\s*$",
               adHoc=lambda p: p.backend.addValue(
                   'relativity_method', FPLO_RELATIVISTIC[p.lastMatch['x_fplo_t_relativity_method']])
            ),
            SM(name='mXC',
               startReStr=r"\s*XC version :\s*(?P<x_fplo_xc_functional_number>\d+)\s*-\s*(?P<x_fplo_xc_functional>.*?)\s*$",
               adHoc=lambda p: self.addSectionDictList(
                   p.backend, 'section_XC_functionals',
                   FPLO_XC_FUNCTIONAL[p.lastMatch['x_fplo_xc_functional']])
            ),
            SM(name='mLSDApU_stuff',
               startReStr=r"\s*LSDA\+U:\s*-+\s*$",
               subMatchers=[
                   SM(name='mLSDApU_projection',
                      startReStr=r"\s*LSDA\+U:\s*Projection\s*:\s*(?P<x_fplo_dft_plus_u_projection_type>\S+?)\s*$",
                      adHoc=lambda p: p.backend.addValue('dft_plus_u_projection_type',
                          FPLO_DFT_PLUS_U_PROJECTION_TYPE[p.lastMatch['x_fplo_dft_plus_u_projection_type']]),
                   ),
                   SM(name='mLSDApU_functional',
                      startReStr=r"\s*LSDA\+U:\s*Functional\s*:\s*(?P<x_fplo_dft_plus_u_functional>.+?)\s*$",
                      adHoc=lambda p: p.backend.addValue('dft_plus_u_functional',
                          FPLO_DFT_PLUS_U_FUNCTIONAL[p.lastMatch['x_fplo_dft_plus_u_functional']]),
                   ),
                   SM(name='mLSDApU_n_correlated_states',
                      startReStr=r"\s*LSDA\+U:\s*\d+\s*Correlated states",
                      subMatchers=[
                          SM(name='mLSDApU_correlated_states_header',
                             startReStr=r'\s*LSDA\+U:\s*sort\s+el\.\s+state\s+F0\s+F2\s+F4\s+F6\s+U\s+J\s*$',
                             subMatchers=[
                                 SM(name='mLSDApU_correlated_states_in_EV',
                                    startReStr=r'\s*LSDA\+U:\s*\(in eV\)\s*$',
                                 ),
                                 SM(name='mLSDApU_correlated_states_line', repeats=True,
                                    startReStr=(
                                        r"\s*LSDA\+U:" +
                                        r"\s*(?P<x_fplo_t_dft_plus_u_species_subshell_species>\d+)" +
                                        r"\s+(?P<x_fplo_t_dft_plus_u_species_subshell_element>\S+)" +
                                        r"\s+(?P<x_fplo_t_dft_plus_u_species_subshell_subshell>\S+)" +
                                        r"\s+(?P<x_fplo_t_dft_plus_u_species_subshell_F0__eV>" + RE_f + r")" +
                                        r"\s+(?P<x_fplo_t_dft_plus_u_species_subshell_F2__eV>" + RE_f + r")" +
                                        r"\s+(?P<x_fplo_t_dft_plus_u_species_subshell_F4__eV>" + RE_f + r")" +
                                        r"\s+(?P<x_fplo_t_dft_plus_u_species_subshell_F6__eV>" + RE_f + r")" +
                                        r"\s+(?P<x_fplo_t_dft_plus_u_species_subshell_U__eV>" + RE_f + r")" +
                                        r"\s+(?P<x_fplo_t_dft_plus_u_species_subshell_J__eV>" + RE_f + r")" +
                                        r"\s*$"
                                    )
                                 ),
                             ],
                          ),
                          SM(name='mLSDApU_correlated_sites_header',
                             startReStr=r'\s*LSDA\+U:\s+site\s+el\.\s+udef\s+state\s+ubi1\s+ubi2\s*$',
                             subMatchers=[
                                 SM(name='mLSDApU_correlated_sites_line', repeats=True,
                                    startReStr=(
                                        r"\s*LSDA\+U:" +
                                        r"\s*(?P<x_fplo_t_dft_plus_u_site_index>\d+)" +
                                        r"\s+(?P<x_fplo_t_dft_plus_u_site_element>\S+)" +
                                        r"\s*(?P<x_fplo_t_dft_plus_u_site_species>\d+)" +
                                        r"\s+(?P<x_fplo_t_dft_plus_u_site_subshell>\S+)" +
                                        r"\s+(?P<x_fplo_t_dft_plus_u_site_ubi1>\d+)" +
                                        r"\s+(?P<x_fplo_t_dft_plus_u_site_ubi2>\d+)" +
                                        r"\s*$"
                                    )
                                 ),
                             ],
                          ),
                      ],
                   ),
               ],
            ),
        ]
        return result

    def SMs_scf(self):
        result = [
            SM(name='new_iter', repeats=True,
               startReStr=r"\s*SCF: iteration\s*(?:\d+)\s*dimension\s*(?:\d+)\s*last deviation\s*u=\s*(?:" + RE_f + r")\s*$",
               sections=['section_scf_iteration'],
               subMatchers=[
                   SM(name='estimate_eFermi',
                      startReStr=r"\s*Estimating Fermi energy\s*\.\.\.\s*$",
                      subMatchers=[
                          SM(name='tetwts_eFermi',
                             startReStr=(
                                r"\s*TETWTS: Fermi energy:\s*(?P<x_fplo_t_energy_reference_fermi_iteration__hartree>" + RE_f + r")\s*;" +
                                r"\s*(?:" + RE_f + r")\s*electrons" +
                                r"\s*$"
                             ),
                          ),
                      ],
                   ),
                   SM(name='header_eTot',
                      startReStr=r"\s*==========\s*TOTAL ENERGY\s*==========\s*$",
                      subMatchers=[
                          SM(name='header2_eTot',
                             startReStr=r"\s*total energy\s+kinetic energy\s+potential energy\s+ex\.-corr\. energy\s*$",
                             subMatchers=[
                                 SM(name='eTot',
                                    startReStr=(
                                        r"EE:" +
                                        r"\s*(?P<energy_total_scf_iteration__eV>" + RE_f + r")" +
                                        r"\s+(?P<electronic_kinetic_energy_scf_iteration__eV>" + RE_f + r")" +
                                        r"\s+(?:" + RE_f + r")" +
                                        r"\s+(?:" + RE_f + r")" +
                                        r"\s*$"
                                    )
                                 ),
                             ],
                          ),
                          SM(name='header2_eTot_DFTU',
                             startReStr=r"\s*total energy\s+kinetic energy\s+potential energy\s+ex\.-corr\. energy\s*"
                                        r"LS(?:AD|DA)\+U energy\s*$",
                             subMatchers=[
                                 SM(name='eTot',
                                    startReStr=(
                                        r"EE:" +
                                        r"\s*(?P<energy_total_scf_iteration__eV>" + RE_f + r")" +
                                        r"\s+(?P<electronic_kinetic_energy_scf_iteration__eV>" + RE_f + r")" +
                                        r"\s+(?:" + RE_f + r")" +
                                        r"\s+(?:" + RE_f + r")" +
                                        r"\s+(?:" + RE_f + r")" +
                                        r"\s*$"
                                    )
                                 ),
                             ],
                          ),
                      ]
                   ),
               ],
            ),
        ]
        return result

    def onOpen_section_scf_iteration(
            self, backend, gIndex, section):
        self.section['scf_iteration'] = section
        self.sectionIdx['scf_iteration'] = gIndex

    def onClose_section_scf_iteration(
            self, backend, gIndex, section):
        eFermi = np.array([
            section['x_fplo_t_energy_reference_fermi_iteration'][-1],
            section['x_fplo_t_energy_reference_fermi_iteration'][-1],
        ])
        backend.addArrayValues('energy_reference_fermi_iteration', eFermi)

    def get_dft_plus_u_per_species_orbital(self, section_method):
        dft_plus_u_per_species_orbital = {}
        for species_orbital_idx in range(len(section_method['x_fplo_t_dft_plus_u_species_subshell_species'])):
            species = section_method['x_fplo_t_dft_plus_u_species_subshell_species'][species_orbital_idx]
            subshell = section_method['x_fplo_t_dft_plus_u_species_subshell_subshell'][species_orbital_idx]
            if dft_plus_u_per_species_orbital.get(species,None) is None:
                dft_plus_u_per_species_orbital[species] = {}
            dft_plus_u_per_species_orbital[species][subshell] = {
                # code-independent orbital data
                'dft_plus_u_orbital_label': subshell,
                'dft_plus_u_orbital_U':
                    section_method['x_fplo_t_dft_plus_u_species_subshell_U'][species_orbital_idx],
                'dft_plus_u_orbital_J':
                    section_method['x_fplo_t_dft_plus_u_species_subshell_J'][species_orbital_idx],
                # code-specific orbital data
                'x_fplo_dft_plus_u_orbital_species': species,
                'x_fplo_dft_plus_u_orbital_element':
                    section_method['x_fplo_t_dft_plus_u_species_subshell_element'][species_orbital_idx],
                'x_fplo_dft_plus_u_orbital_F0':
                    section_method['x_fplo_t_dft_plus_u_species_subshell_F0'][species_orbital_idx],
                'x_fplo_dft_plus_u_orbital_F2':
                    section_method['x_fplo_t_dft_plus_u_species_subshell_F2'][species_orbital_idx],
                'x_fplo_dft_plus_u_orbital_F4':
                    section_method['x_fplo_t_dft_plus_u_species_subshell_F4'][species_orbital_idx],
                'x_fplo_dft_plus_u_orbital_F6':
                    section_method['x_fplo_t_dft_plus_u_species_subshell_F6'][species_orbital_idx],
            }
        return dft_plus_u_per_species_orbital

    def get_dft_plus_u_orbitals(self, section_method):
        dft_plus_u_orbitals = []
        dft_plus_u_per_species_orbital = self.get_dft_plus_u_per_species_orbital(section_method)
        for dft_plus_u_idx in range(len(section_method['x_fplo_t_dft_plus_u_site_index'])):
            site = section_method['x_fplo_t_dft_plus_u_site_index'][dft_plus_u_idx]
            element = section_method['x_fplo_t_dft_plus_u_site_element'][dft_plus_u_idx]
            species = section_method['x_fplo_t_dft_plus_u_site_species'][dft_plus_u_idx]
            orbital = section_method['x_fplo_t_dft_plus_u_site_subshell'][dft_plus_u_idx]
            dft_plus_u_orbital = dft_plus_u_per_species_orbital[species][orbital].copy()
            dft_plus_u_orbital['dft_plus_u_orbital_atom'] = site - 1 # FPLO site counts one-based
            dft_plus_u_orbitals.append(dft_plus_u_orbital)
        return dft_plus_u_orbitals

    def onOpen_section_method(
            self, backend, gIndex, section):
        self.sectionIdx['method'] = gIndex
        self.section['method'] = section

    def onClose_section_method(
            self, backend, gIndex, section):
        # check for DFT+U vs. DFT
        if section['x_fplo_dft_plus_u_projection_type'] is None:
            backend.addValue('electronic_structure_method', 'DFT')
        else:
            backend.addValue('electronic_structure_method', 'DFT+U')
            self.addSectionDictList(backend, 'section_dft_plus_u_orbital',
                                    self.get_dft_plus_u_orbitals(section))

    def onOpen_section_run(
            self, backend, gIndex, section):
        self.section['run'] = section
        self.sectionIdx['run'] = gIndex

    def onClose_section_run(
            self, backend, gIndex, section):
        # assemble version number
        backend.addValue('program_version',
                         section['x_fplo_t_program_version_main'][-1] + '-' +
                         section['x_fplo_t_program_version_release'][-1])
        # map list of hosts to dict
        backend.addValue('run_hosts',
                         {h:1 for h in section['x_fplo_t_run_hosts']})

    def onOpen_section_system(
            self, backend, gIndex, section):
        self.section['system'] = section
        self.sectionIdx['system'] = gIndex

    def onClose_section_system(
            self, backend, gIndex, section):
        # store direct lattice matrix and inverse for transformation crystal <-> cartesian
        if section['x_fplo_t_vec_a_x'] is not None:
            self.amat = np.array([
                section['x_fplo_t_vec_a_x'][-3:], section['x_fplo_t_vec_a_y'][-3:], section['x_fplo_t_vec_a_z'][-3:],
            ], dtype=np.float64).T
            # store inverse for transformation cartesian -> crystal
            try:
                self.amat_inv = np.linalg.inv(self.amat)
            except np.linalg.linalg.LinAlgError:
                raise Exception("error inverting bravais matrix " + str(self.amat))
            LOGGER.info('NewCell')
        else:
            raise Exception("missing bravais vectors")
        backend.addArrayValues('simulation_cell', self.amat)
        # store reciprocal lattice matrix and inverse for transformation crystal <-> cartesian
        if section['x_fplo_t_vec_b_x'] is not None:
            self.bmat = np.array([
                section['x_fplo_t_vec_b_x'], section['x_fplo_t_vec_b_y'], section['x_fplo_t_vec_b_z'],
            ], dtype=np.float64).T
            # store inverse for transformation cartesian -> crystal
            try:
                self.bmat_inv = np.linalg.inv(self.bmat)
            except np.linalg.linalg.LinAlgError:
                raise Exception("error inverting reciprocal cell matrix")
        elif section['x_fplo_t_vec_a_x'] is not None:
            # we got new lattice vectors, but no reciprocal ones, calculate
            # on-the-fly
            LOGGER.error('calculating bmat on the fly from amat')
            abmat = np.zeros((3,3), dtype=np.float64)
            abmat[0] = np.cross(self.amat[1],self.amat[2])
            abmat[1] = np.cross(self.amat[2],self.amat[0])
            abmat[2] = np.cross(self.amat[0],self.amat[1])
            abmat *= 2*math.pi / np.dot(abmat[0],self.amat[0])
            self.bmat = abmat
            # store inverse for transformation cartesian -> crystal
            try:
                self.bmat_inv = np.linalg.inv(self.bmat)
            except np.linalg.linalg.LinAlgError:
                raise Exception("error inverting reciprocal cell matrix")
        else:
            raise Exception("missing reciprocal cell vectors")
        backend.addArrayValues('x_fplo_reciprocal_cell', self.bmat)
        # stuff we parsed from atom position table
        if section['x_fplo_t_atom_labels'] is not None:
            backend.addArrayValues('atom_labels', np.asarray(
                section['x_fplo_t_atom_labels']))
        if section['x_fplo_t_atom_idx'] is not None:
            backend.addArrayValues('x_fplo_atom_idx', np.asarray(
                section['x_fplo_t_atom_idx']))
        if section['x_fplo_t_atom_wyckoff_idx'] is not None:
            backend.addArrayValues('x_fplo_atom_wyckoff_idx', np.asarray(
                section['x_fplo_t_atom_wyckoff_idx']))
        if section['x_fplo_t_atom_cpa_block'] is not None:
            backend.addArrayValues('x_fplo_atom_cpa_block', np.asarray(
                section['x_fplo_t_atom_cpa_block']))
        if section['x_fplo_t_atom_positions_x'] is not None:
            backend.addArrayValues('atom_positions', np.asarray([
                section['x_fplo_t_atom_positions_x'],
                section['x_fplo_t_atom_positions_y'],
                section['x_fplo_t_atom_positions_z'],
            ], dtype=np.float64).T)

    def onOpen_section_single_configuration_calculation(
            self, backend, gIndex, section):
        self.section['single_configuration_calculation'] = section
        self.sectionIdx['single_configuration_calculation'] = gIndex

    def onClose_section_single_configuration_calculation(
            self, backend, gIndex, section):
        backend.addValue('single_configuration_calculation_to_system_ref', self.sectionIdx['system'])
        backend.addValue('single_configuration_to_calculation_method_ref', self.sectionIdx['method'])
        scf_iter = self.section['scf_iteration']
        backend.addValue(
            'energy_total',
            scf_iter['energy_total_scf_iteration'][-1]
        )
        backend.addArrayValues(
            'energy_reference_fermi',
            np.array(scf_iter['energy_reference_fermi_iteration'][-1])
        )

    def initialize_values(self):
        """allows to reset values if the same superContext is used to parse
        different files"""
        self.section = {}
        self.sectionIdx = {}

    def startedParsing(self, path, parser):
        """called when parsing starts"""
        self.parser = parser
        # reset values if same superContext is used to parse different files
        self.initialize_values()

    def strValueTransform_strFploDate(self, fplo_date):
        if fplo_date is None:
            return None
        epoch = 0
        match = re.match(
            r"(?P<dow>[A-Za-z]+)\s+(?P<month>[A-Za-z]+)\s+(?P<day>\d+)\s+" +
            r"(?P<hour>\d+):\s*(?P<minute>\d+):\s*(?P<second>\d+)\s+" +
            r"(?P<year>\d+)\s*",
            fplo_date)
        if match:
            month = FploC.MONTH_NUMBER[match.group('month')]
            epoch = calendar.timegm(
                (int(match.group('year')), int(month), int(match.group('day')),
                 int(match.group('hour')), int(match.group('minute')), int(match.group('second'))))
        else:
            raise RuntimeError("unparsable date: %s", fplo_date)
        return(epoch)
    strValueTransform_strFploDate.units = 's'

    def addSectionDictList(self, backend, section_name, section_dict_list):
        for section_dict in section_dict_list:
            self.addSectionDict(backend, section_name, section_dict)

    def addSectionDict(self, backend, section_name, section_dict):
        gIndex = backend.openSection(section_name)
        for key, value in sorted(section_dict.items()):
            if isinstance(value, (list,dict)):
                backend.addValue(key, value)
            else:
                backend.addValue(key, value)
        backend.closeSection(section_name, gIndex)


class FploParser():
    """ A proper class envolop for running this parser from within python. """
    def __init__(self, backend, **kwargs):
        self.backend_factory = backend

    def parse(self, mainfile):
        from unittest.mock import patch
        logging.info('fplo parser started')
        logging.getLogger('nomadcore').setLevel(logging.WARNING)
        backend = self.backend_factory("fplo.nomadmetainfo.json")
        parserInfo = {'name': 'fplo-parser', 'version': '1.0'}
        context = ParserFplo14(backend.metaInfoEnv())

        with patch.object(sys, 'argv', ['<exe>', '--uri', 'nmd://uri', mainfile]):
            mainFunction(
                context.mainFileDescription(),
                None,
                parserInfo=FploC.PARSER_INFO_DEFAULT.copy(),
                cachingLevelForMetaName=context.cachingLevelForMetaName,
                superContext=context,
                superBackend=backend)

        return backend


from types import ModuleType
import sys
from importlib._bootstrap import _ImportLockContext
from six import raise_from
from importlib import reload as reload_module


__all__ = ['lazy_module', 'LazyModule', '_MSG']


_CLS_ATTRS = (
    '_lazy_import_error_strings', '_lazy_import_error_msgs', '_lazy_import_callables',
    '_lazy_import_submodules', '__repr__'
)

_DICT_DELETION = ('_lazy_import_submodules',)

_MSG = ("{caller} attempted to use a functionality that requires module "
        "{module}, but it couldn't be loaded. Please install {install_name} "
        "and retry.")


class LazyModule(ModuleType):
    def __getattribute__(self, attr):
        if attr not in ('__name__', '__class__', '__spec__'):
            try:
                name = '%s.%s' % (self.__name__, attr)
                return sys.modules[name]
            except KeyError:
                pass

            try:
                return type(self)._lazy_import_callables[attr]
            except (AttributeError, KeyError):
                _load_module(self)
        return super(LazyModule, self).__getattribute__(attr)

    def __setattr__(self, attr, value):
        _load_module(self)
        return super(LazyModule, self).__setattr__(attr, value)


def _clean_lazy_submodule_refs(module):
    module_class = type(module)
    for entry in _DICT_DELETION:
        try:
            names = getattr(module_class, entry)
        except AttributeError:
            continue
        for name in names:
            try:
                super(LazyModule, module).__delattr__(name)
            except AttributeError:
                pass


def _clean_lazymodule(module):
    module_class = type(module)
    _clean_lazy_submodule_refs(module)

    module_class.__getattribute__ = ModuleType.__getattribute__
    module_class.__setattr__ = ModuleType.__setattr__
    class_attrs = {}
    for attr in _CLS_ATTRS:
        try:
            class_attrs[attr] = getattr(module_class, attr)
            delattr(module_class, attr)
        except AttributeError:
            pass
    return class_attrs


def _reset_lazy_submodule_refs(module):
    module_class = type(module)
    for entry in _DICT_DELETION:
        try:
            names = getattr(module_class, entry)
        except AttributeError:
            continue
        for name, submodule in names.items():
            super(LazyModule, module).__setattr__(name, submodule)


def _reset_lazymodule(module, class_attrs):
    module_class = type(module)
    del module_class.__getattribute__
    del module_class.__setattr__
    try:
        del module_class._LOADING
    except AttributeError:
        pass

    for attr in _CLS_ATTRS:
        try:
            setattr(module_class, attr, class_attrs[attr])
        except KeyError:
            pass

    _reset_lazy_submodule_refs(module)


def _load_module(module):
    module_class = type(module)
    if not issubclass(module_class, LazyModule):
        raise TypeError('Not an instance of LazyModule')
    with _ImportLockContext():
        parent, _, module_name = module.__name__.rpartition('.')
        if not hasattr(module_class, '_lazy_import_error_msgs'):
            return
        module_class._LOADING = True
        try:
            if parent:
                setattr(sys.modules[parent], module_name, module)
            if not hasattr(module_class, '_LOADING'):
                return
            cached_data = _clean_lazymodule(module)
            try:
                reload_module(module)
            except Exception:
                _reset_lazymodule(module, cached_data)
                raise
            else:
                delattr(module_class, '_LOADING')
                _reset_lazy_submodule_refs(module)
        except (AttributeError, ImportError):
            msg = module_class._lazy_import_error_msgs['msg']
            raise_from(ImportError(msg.format(**module_class._lazy_import_error_strings)), None)


def _lazy_module(module_name, error_strings):
    with _ImportLockContext():
        full_module_name = module_name
        full_submodule_name = None
        submodule_name = ''
        while module_name:
            try:
                module = sys.modules[module_name]
                module_name = ''
            except KeyError:
                err_strs = error_strings.copy()
                err_strs.setdefault('module', module_name)

                class _LazyModule(LazyModule):
                    _lazy_import_error_msgs = {'msg': err_strs.pop('msg')}
                    msg_callable = err_strs.pop('msg_callable', None)
                    if msg_callable:
                        _lazy_import_error_msgs['msg_callable'] = msg_callable
                    _lazy_import_error_strings = err_strs
                    _lazy_import_callables = {}
                    _lazy_import_submodules = {}

                    def __repr__(self):
                        return 'Lazily-loaded module %s' % self.__name__

                _LazyModule.__name__ = 'module'
                module = sys.modules[module_name] = _LazyModule(module_name)

            if full_submodule_name:
                submodule = sys.modules[full_submodule_name]
                ModuleType.__setattr__(module, submodule_name, submodule)
                _LazyModule._lazy_import_submodules[submodule_name] = submodule

            full_submodule_name = module_name
            module_name, _, submodule_name = module_name.rpartition('.')

        return sys.modules[full_module_name]


def lazy_module(module_name, level='leaf'):
    module_base_name = module_name.partition('.')[0]
    error_strings = {}
    try:
        caller = sys._getframe(3).f_globals['__name__']
    except AttributeError:
        caller = 'Python'
    error_strings.setdefault('caller', caller)
    error_strings.setdefault('install_name', module_base_name)
    error_strings.setdefault('msg', _MSG)

    module = _lazy_module(module_name, error_strings)

    if level == 'base':
        return sys.modules[module_base_name]
    elif level == 'leaf':
        return module
    else:
        raise ValueError('Must be base or leaf')



from lark import Transformer


class DebugTransformer(Transformer):  # pragma: no cover
    def __init__(self):
        super().__init__()

    def __default__(self, data, children, meta):
        print("Node: ", data, children)
        return data


class TransformerSkeleton(Transformer):  # pragma: no cover
    """Prints out all the nodes and its arguments during the walk-through of the tree."""

    def __init__(self):
        super().__init__()

    def filter(self, arg):
        # filter: expression*
        print("Node: ", "filter", arg)
        return "filter"

    def constant(self, arg):
        # constant: string | number
        print("Node: ", "constant", arg)
        return "constant"

    def value(self, arg):
        # value: string | number | property
        print("Node: ", "value", arg)
        return "value"

    def value_list(self, arg):
        # value_list: [ OPERATOR ] value ( "," [ OPERATOR ] value )*
        print("Node: ", "value_list", arg)
        return "value_list"

    def value_zip(self, arg):
        # value_zip: [ OPERATOR ] value ":" [ OPERATOR ] value (":" [ OPERATOR ] value)*
        print("Node: ", "value_zip", arg)
        return "value_zip"

    def value_zip_list(self, arg):
        # value_zip_list: value_zip ( "," value_zip )*
        print("Node: ", "value_zip_list", arg)
        return "value_zip_list"

    def expression(self, arg):
        # expression: expression_clause [ OR expression ]
        print("Node: ", "expression", arg)
        return "expression"

    def expression_clause(self, arg):
        # expression_clause: expression_phrase [ AND expression_clause ]
        print("Node: ", "expression_clause", arg)
        return "expression_clause"

    def expression_phrase(self, arg):
        # expression_phrase: [ NOT ] ( comparison | predicate_comparison | "(" expression ")" )
        print("Node: ", "expression_phrase", arg)
        return "expression_phrase"

    def comparison(self, arg):
        # comparison: constant_first_comparison | property_first_comparison
        # Note: Do nothing!
        print("Node: ", "comparison", arg)
        return "comparison"

    def property_first_comparison(self, arg):
        # property_first_comparison: property ( value_op_rhs | known_op_rhs | fuzzy_string_op_rhs | set_op_rhs |
        # set_zip_op_rhs )
        print("Node: ", "property_first_comparison", arg)
        return "property_first_comparison"

    def constant_first_comparison(self, arg):
        # constant_first_comparison: constant value_op_rhs
        print("Node: ", "constant_first_comparison", arg)
        return "constant_first_comparison"

    def predicate_comparison(self, arg):
        # predicate_comparison: length_comparison
        print("Node: ", "predicate_comparison", arg)
        return "predicate_comparison"

    def value_op_rhs(self, arg):
        # value_op_rhs: OPERATOR value
        print("Node: ", "value_op_rhs", arg)
        return "value_op_rhs"

    def known_op_rhs(self, arg):
        # known_op_rhs: IS ( KNOWN | UNKNOWN )
        print("Node: ", "known_op_rhs", arg)
        return "known_op_rhs"

    def fuzzy_string_op_rhs(self, arg):
        # fuzzy_string_op_rhs: CONTAINS string | STARTS [ WITH ] string | ENDS [ WITH ] string
        print("Node: ", "fuzzy_string_op_rhs", arg)
        return "fuzzy_string_op_rhs"

    def set_op_rhs(self, arg):
        # set_op_rhs: HAS ( [ OPERATOR ] value | ALL value_list | ANY value_list | ONLY value_list )
        print("Node: ", "set_op_rhs", arg)
        return "set_op_rhs"

    def set_zip_op_rhs(self, arg):
        # set_zip_op_rhs: property_zip_addon HAS ( value_zip | ONLY value_zip_list | ALL value_zip_list |
        # ANY value_zip_list )
        print("Node: ", "set_zip_op_rhs", arg)
        return "set_zip_op_rhs"

    def length_comparison(self, arg):
        # length_comparison: LENGTH property OPERATOR value
        print("Node: ", "length_comparison", arg)
        return "length_comparison"

    def property_zip_addon(self, arg):
        # property_zip_addon: ":" property (":" property)*
        print("Node: ", "property_zip_addon", arg)
        return "property_zip_addon"

    def property(self, arg):
        # property: IDENTIFIER ( "." IDENTIFIER )*
        print("Node: ", "property", arg)
        return "property"

    def string(self, arg):
        # string: ESCAPED_STRING
        print("Node: ", "string", arg)
        return "string"

    def number(self, arg):
        # number: SIGNED_INT | SIGNED_FLOAT
        print("Node: ", "number", arg)
        return "number"

    def __default__(self, data, children, meta):
        print("Node: ", data, children)
        return data


if __name__ == "__main__":  # pragma: no cover
    from optimade.filterparser import LarkParser

    p = LarkParser(version=(0, 10, 0))
    # t = DebugTransformer()
    t = TransformerSkeleton()

    # f = 'a.a = "text" OR a<a AND NOT b>=8'

    # single list

    f = "list HAS < 3"

    f = "list HAS < 3, > 4"  # -> error
    f = "list HAS ALL < 3, > 4"

    # multiple lists

    f = "list1:list2 HAS < 3 : > 4"
    f = "list1:list2 HAS ALL < 3 : > 4"

    f = "list1:list2 HAS < 3 : > 4, < 2 : > 5"  # -> error
    f = "list1:list2 HAS ALL < 3 : > 4, < 2 : > 5"
    f = "list1:list2 HAS ALL < 3, < 2 : > 4, > 5"  # -> error

    # f = 'list1:list2 HAS < 3, > 4'  # -> error
    # f = 'list1:list2 HAS ALL < 3, > 4'  # -> error

    f = 'elements:elements_ratios HAS ALL "Al":>0.3333, "Al":<0.3334'
    f = 'elements:elements_ratios HAS ALL "Al":>0.3333 AND elements_ratio<0.3334'
    f = 'elements:elements_ratios HAS ALL "Al" : >0.3333, <0.3334'  # -> error

    f = "list1:list2 HAS ALL < 3 : > 4, < 2 : > 5 : > 4, < 2 : > 5"  # valid but wrong
    f = "ghf.flk<gh"  # valid but wrong

    # f = ''

    tree = p.parse(f)
    print(tree)
    print(tree.pretty())

    t.transform(tree)



# Copyright 2016-2018 Fawzi Mohamed, Carl Poelking, Daria Tomecka
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import print_function
from builtins import zip
from builtins import str
from builtins import map
from builtins import range
from builtins import object
import os
import sys
import re
import numpy as np
import xml.dom.minidom
from nomad.utils import create_uuid

try:
    import ase
    import ase.io
    HAVE_ASE = True
except ImportError:
    HAVE_ASE = False
    pass

def XmlGetUnique(root, tag):
    nodes = XmlGetAll(root, tag)
    if len(nodes) > 1:
        raise ValueError("More than one node with tag '%s'" % tag)
    return nodes[0]

def XmlGetAll(root, tag):
    nodes = root.getElementsByTagName(tag)
    ret_nodes = []
    for node in nodes:
        if node.localName != None: ret_nodes.append(node)
    return ret_nodes

def XmlGetChildDict(root):
    tag_child = {}
    for child in root.childNodes:
        if child.localName == None: continue
        if not child.localName in tag_child:
            tag_child[child.localName] = []
        tag_child[child.localName].append(child)
    return tag_child

def XmlGetAttributes(root):
    return root.attributes

def XmlGetText(root):
    return root.firstChild.nodeValue



class LibAtomsParser(object):
    def __init__(self, log=None):
        self.log = log
        self.data = {}
        self.logtag = 'main'
        # KEY DEFAULT DICTIONARIES
        self.missing_keys_lh = [] # Transform keys that were not found in output
        self.missing_keys_rh = []
        self.ignored_keys = [] # Raw keys that did not have a transform
        self.keys_not_found = [] # Searches that failed
        return
    def __getitem__(self, key):
        self.selected_data_item = self.data[key]
        return self
    def As(self, typ=None):
        if typ == None:
            return self.selected_data_item
        return typ(self.selected_data_item)
    def SummarizeKeyDefaults(self):
        if not self.log: return
        if len(self.missing_keys_lh):
            self.log << self.log.my \
                << "[%s] Keys from transformation maps that went unused (=> set to 'None'):" \
                % self.logtag << self.log.endl
            for lh, rh in zip(self.missing_keys_lh, self.missing_keys_rh):
                self.log << self.log.item << "Key = %-25s <> %25s" % (rh, lh) << self.log.endl
        if len(self.ignored_keys):
            self.log << self.log.mb \
                << "[%s] Keys from XY mapping that were not transformed (=> not stored):" \
                % self.logtag << self.log.endl
            for key in self.ignored_keys:
                self.log << self.log.item << "Key =" << key << self.log.endl
        if len(self.keys_not_found):
            self.log << self.log.mr \
                << "[%s] Keys from searches that failed (=> set to 'None'):" \
                % self.logtag << self.log.endl
            for key in self.keys_not_found:
                self.log << self.log.item << "Key =" << key << self.log.endl
        return
    def Set(self, key, value):
        if self.log:
            self.log << "Set [%s]   %-40s = %s" % (self.logtag, key, str(value)) << self.log.endl
        if key not in self.data:
            self.data[key] = value
        else:
            raise KeyError("Key already exists: '%s'" % key)
        return
    def SearchMapKeys(self, expr, ln, keys):
        s = re.search(expr, ln)
        try:
            for i in range(len(keys)):
                self.Set(keys[i], s.group(i+1).strip())
        except AttributeError:
            for i in range(len(keys)):
                self.Set(keys[i], None)
                self.keys_not_found.append(keys[i])
        return
    def ReadBlockXy(self, block):
        lns = block.lns
        block_data = {}
        for ln in lns:
            ln = ln.replace('\n','')
            if ln == '':
                continue
            if ':' in ln:
                sp = ln.split(':')
                x = sp[0].strip().split()
                y = sp[1].strip()
            elif '=' in ln:
                sp = ln.split('=')
                x = sp[0].strip().split()
                y = sp[1].strip()
            else:
                sp = ln.split()
                x = sp[:-1]
                y = sp[-1]
            key = ''
            for i in range(len(x)-1):
                xi = x[i].replace('(','').replace(')','').lower()
                key += '%s_' % xi
            key += '%s' % x[-1].replace('(','').replace(')','').lower()
            value = y
            block_data[key] = value
        return block_data
    def ApplyBlockXyData(self, block_data, key_map):
        for key_in in key_map:
            key_out = key_map[key_in]
            if key_in not in block_data:
                # Missing key in output
                self.missing_keys_lh.append(key_in)
                self.missing_keys_rh.append(key_out)
                value = None
            else:
                value = block_data[key_in]
            if key_out == None:
                key_out = key_in
            self.Set(key_out, value)
        for key in block_data:
            if key not in key_map:
                # Missing key in transform map
                self.ignored_keys.append(key)
        return
    def ParseOutput(self, output_file):
        if self.log:
            self.log << self.log.mg << "libAtomsParser::ParseOutput ..." << self.log.endl

        if HAVE_ASE:
            read_fct = ase.io.read
            read_fct_args = { 'index':':' }
        else:
            raise NotImplementedError("None-ASE read function requested, but not yet available.")
            read_fct = None
            read_fct_args = None

        # PARSE CONFIGURATIONS
        self.ase_configs = read_fct(output_file, **read_fct_args)
        # for config in ase_configs:
        #     print(config)

        self.Set('program_name', 'libAtoms')
        self.Set('program_version', 'n/a')
        return

class LibAtomsGapParser(LibAtomsParser):
    def __init__(self, log=None):
        super(LibAtomsGapParser, self).__init__(log)
        self.logtag = 'gap-xml'
        self.trj = None
        self.has_gap_data = False
        return
    def ParseOutput(self, output_file, base_dir=''):
        self.Set('program_name', 'libAtoms')
        dom = xml.dom.minidom.parse(output_file)
        root = XmlGetUnique(dom, 'GAP_params')
        # Child keys should be: ['gpSparse', 'command_line', 'GAP_data', 'XYZ_data']
        # print(list(child_nodes.keys()))
        child_nodes = XmlGetChildDict(root)

        # 'GAP_params'
        atts = XmlGetAttributes(root)
        keys = ['label', 'svn_version']
        for key in keys:
            self.Set('GAP_params.%s' % key, atts[key].value) # TODO Handle look-up errors

        # 'GAP_params/GAP_data'
        key = 'GAP_data'
        if key in child_nodes:
            node = child_nodes[key][0]
            atts = XmlGetAttributes(node)
            keys = ['do_core', 'e0']
            for key in keys:
                self.Set('GAP_data.%s' % key, atts[key].value) # TODO Handle look-up errors

        # 'GAP_params/command_line'
        key = 'command_line'
        if key in child_nodes:
            node = child_nodes[key][0]
            text = XmlGetText(node)
            self.Set('command_line.command_line', text)

        # 'GAP_params/gpSparse'
        key = 'gpSparse'
        if key in child_nodes:
            self.has_gap_data = True
            node = child_nodes[key][0]
            atts = XmlGetAttributes(node)
            keys = ['n_coordinate']
            for key in keys:
                self.Set('gpSparse.%s' % key, atts[key].value) # TODO Handle look-up errors

            # GAP_params/gpSparse/gpCoordinates
            gp_coord_node = XmlGetUnique(node, 'gpCoordinates')
            gp_coord_child_nodes = XmlGetChildDict(gp_coord_node)
            gp_coord_node_att = XmlGetAttributes(gp_coord_node)
            for key in gp_coord_node_att.keys():
                self.Set('gpCoordinates.%s' % key, gp_coord_node_att[key].value)

            # 'GAP_params/gpSparse/gpCoordinates/theta
            key = 'theta'
            if key in gp_coord_child_nodes:
                node = gp_coord_child_nodes[key][0]
                text = XmlGetText(node).strip()
                self.Set('gpCoordinates.%s' % key, text)
            # 'GAP_params/gpSparse/gpCoordinates/descriptor
            key = 'descriptor'
            if key in gp_coord_child_nodes:
                node = gp_coord_child_nodes[key][0]
                text = XmlGetText(node).strip()
                self.Set('gpCoordinates.%s' % key, text)
            # 'GAP_params/gpSparse/gpCoordinates/descriptor
            key = 'permutation'
            if key in gp_coord_child_nodes:
                node = gp_coord_child_nodes[key][0]
                att = XmlGetAttributes(node)
                text = XmlGetText(node).strip()
                self.Set('gpCoordinates.perm.%s' % key, text)
                self.Set('gpCoordinates.perm.i', att['i'].value)
            # 'GAP_params/gpSparse/gpCoordinates/sparseX
            key = 'sparseX'
            if key in gp_coord_child_nodes:
                n_sparseX = self['gpCoordinates.n_sparseX'].As(int)
                n_dim = self['gpCoordinates.dimensions'].As(int)
                sparseX_filename = self['gpCoordinates.sparseX_filename'].As(str)
                sparseX_filename = os.path.join(base_dir, sparseX_filename)
                # Read alpha coefficients
                nodes = gp_coord_child_nodes[key]
                alpha_cutoff = np.zeros((n_sparseX, 2), dtype='float64')
                for i,node in enumerate(nodes):
                    att = XmlGetAttributes(node)
                    alpha_cutoff[i,0] = float(att["alpha"].value)
                    alpha_cutoff[i,1] = float(att["sparseCutoff"].value)
                self.Set('gpCoordinates.alpha', alpha_cutoff)
                # Read descriptor matrix
                X_matrix = np.loadtxt(sparseX_filename)
                assert X_matrix.shape[0] == n_sparseX*n_dim
                # i-th row of X-matrix stores X-vector of config i
                X_matrix = X_matrix.reshape((n_dim, n_sparseX))
                self.Set('gpCoordinates.sparseX', X_matrix)

        # 'GAP_params/XYZ_data'
        key = 'XYZ_data'
        if key in child_nodes:
            node = child_nodes[key][0]
            text = XmlGetText(node)
            unique_ID = create_uuid()
            trj_file = '/tmp/' + unique_ID + '-lib-atoms-gap.from-xml.xyz'
            ofs = open(trj_file, 'w')
            for child in node.childNodes:
                if child.nodeValue == None: continue
                ln = child.nodeValue.strip(' \n')
                if ln == '': continue
                ofs.write(ln+'\n')
            ofs.close()
            self.trj = LibAtomsTrajectory(self.log)
            self.trj.ParseOutput(trj_file)
        return

class LibAtomsTrajectory(LibAtomsParser):
    def __init__(self, log=None):
        super(LibAtomsTrajectory, self).__init__(log)
        self.ase_configs = None
        self.frames = []
        self.logtag = 'trj'
    def ParseOutput(self, output_file):
        if self.log:
            self.log << self.log.mg << "libAtomsParser::ParseOutput ..." << self.log.endl
        if HAVE_ASE:
            read_fct = ase.io.read
            read_fct_args = { 'index':':' }
        else:
            raise NotImplementedError("None-ASE read function requested, but not yet available.")
            read_fct = None
            read_fct_args = None
        # PARSE CONFIGURATIONS
        self.ase_configs = read_fct(output_file, **read_fct_args)
        self.LoadAseConfigs(self.ase_configs)
        self.Set('program_name', 'libAtoms')
        self.Set('program_version', 'n/a')
        return
    def LoadAseConfigs(self, ase_configs):
        for config in ase_configs:
            frame = LibAtomsFrame(self.log)
            frame.LoadAseConfig(config)
            self.frames.append(frame)
        if self.log: self.log << "Loaded %d configurations" % len(self.frames) << self.log.endl
        return

class LibAtomsFrame(LibAtomsParser):
    def __init__(self, log=None):
        super(LibAtomsFrame, self).__init__(log)
        self.ase_config = None
        self.has_energy = False
        self.energy = None
        self.has_virial = False
        self.virial = None
        self.has_config_type = False
        self.config_type = None
    def LoadAseConfig(self, ase_config):
        self.ase_config = ase_config
        # print("INFO", self.ase_config.info)
        key = 'energy'
        if key in self.ase_config.info:
            self.has_energy = True
            self.energy = self.ase_config.info[key]
        else:
            self.has_energy = True
            self.energy = self.ase_config.get_total_energy()
        key = 'config_type'
        if key in self.ase_config.info:
            self.has_config_type = True
            self.config_type = self.ase_config.info[key]
        key = 'virial'
        if key in self.ase_config.info:
            self.has_virial = True
            self.virial = np.array(self.ase_config.info[key])
        return

# ===================
# FILE & BLOCK STREAM
# ===================

class FileStream(object):
    def __init__(self, filename=None):
        if filename:
            self.ifs = open(filename, 'r')
        else:
            self.ifs = None
        return
    def SkipTo(self, expr):
        while True:
            ln = self.ifs.readline()
            if expr in ln:
                break
            if self.all_read():
                break
        return ln
    def SkipToMatch(self, expr):
        while True:
            ln = self.ifs.readline()
            m = re.search(expr, ln)
            if m:
                return ln
            if self.all_read(): break
        return None
    def GetBlock(self, expr1, expr2):
        inside = False
        outside = False
        block = ''
        block_stream = BlockStream()
        while True:
            last_pos = self.ifs.tell()
            ln = self.ifs.readline()
            if expr1 in ln: inside = True
            if expr2 in ln: outside = True
            if inside and not outside:
                # Inside the block
                block += ln
                block_stream.append(ln)
            elif inside and outside:
                self.ifs.seek(last_pos)
                # Block finished
                break
            else:
                # Block not started yet
                pass
            if self.all_read(): break
        return block_stream
    def GetBlockSequence(self,
            expr_start,
            expr_new,
            expr_end,
            remove_eol=True,
            skip_empty=True):
        inside = False
        outside = False
        # Setup dictionary to collect blocks
        blocks = { expr_start : [] }
        for e in expr_new:
            blocks[e] = []
        # Assume structure like this (i <> inside, o <> outside)
        # Lines with 'i' get "eaten"
        # """
        # o ...
        # i <expr_start>
        # i ...
        # i <expr_new[1]>
        # i ...
        # i <expr_new[0]>
        # i ...
        # o <expr_end>
        # o ...
        # """
        key = None
        while True:
            # Log line position
            last_pos = self.ifs.tell()
            ln = self.ifs.readline()
            # Figure out where we are
            if not inside and expr_start in ln:
                #print "Enter", expr_start
                inside = True
                key = expr_start
                new_block = BlockStream(key)
                blocks[key].append(new_block)
            for expr in expr_new:
                if inside and expr in ln:
                    #print "Enter", expr
                    key = expr
                    new_block = BlockStream(key)
                    blocks[key].append(new_block)
            if inside and expr_end != None and expr_end in ln:
                outside = True
            if inside and not outside:
                # Inside a block
                if remove_eol: ln = ln.replace('\n', '')
                if skip_empty and ln == '': pass
                else: blocks[key][-1].append(ln)
            elif inside and outside:
                # All blocks finished
                self.ifs.seek(last_pos)
                break
            else:
                # No blocks started yet
                pass
            if self.all_read(): break
        return blocks
    def all_read(self):
        return self.ifs.tell() == os.fstat(self.ifs.fileno()).st_size
    def readline(self):
        return ifs.readline()
    def close(self):
        self.ifs.close()
    def nextline(self):
        while True:
            ln = self.ifs.readline()
            if ln.strip() != '':
                return ln
            else: pass
            if self.all_read(): break
        return ln
    def ln(self):
        return self.nextline()
    def sp(self):
        return self.ln().split()
    def skip(self, n):
        for i in range(n):
            self.ln()
        return

class BlockStream(FileStream):
    def __init__(self, label=None):
        super(BlockStream, self).__init__(None)
        self.ifs = self
        self.lns = []
        self.idx = 0
        self.label = label
    def append(self, ln):
        self.lns.append(ln)
    def readline(self):
        if self.all_read():
            return ''
        ln = self.lns[self.idx]
        self.idx += 1
        return ln
    def all_read(self):
        return self.idx > len(self.lns)-1
    def tell(self):
        return self.idx
    def cat(self, remove_eol=True, add_eol=False):
        cat = ''
        for ln in self.lns:
            if remove_eol:
                cat += ln.replace('\n', '')
            elif add_eol:
                cat += ln+'\n'
            else:
                cat += ln
        return cat



# Copyright 2016-2018 Lauri Himanen, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import absolute_import
import re
import logging
import numpy as np
from .commonparser import CPMDCommonParser
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.baseclasses import MainHierarchicalParser
import nomadcore.csvparsing
import nomadcore.configurationreading
LOGGER = logging.getLogger("nomad")


class CPMDMDParser(MainHierarchicalParser):
    """The main parser class that is called for all run types. Parses the CPMD
    output file.
    """
    def __init__(self, parser_context):
        """
        """
        super(CPMDMDParser, self).__init__(parser_context)
        self.setup_common_matcher(CPMDCommonParser(parser_context))
        self.sampling_method_gid = None
        self.frame_refs = []
        self.energies = []

        # Detect what files are available in the same folder where the main
        # file is located.
        self.dcd_filepath = self.file_service.get_absolute_path_to_file("TRAJEC.dcd")
        self.xyz_filepath = self.file_service.get_absolute_path_to_file("TRAJEC.xyz")
        self.trajectory_filepath = self.file_service.get_absolute_path_to_file("TRAJECTORY")
        self.ftrajectory_filepath = self.file_service.get_absolute_path_to_file("FTRAJECTORY")
        self.energies_filepath = self.file_service.get_absolute_path_to_file("ENERGIES")

        #=======================================================================
        # Cache levels
        # self.caching_levels.update({
            # 'section_run': CachingLevel.ForwardAndCache,
        # })

        #=======================================================================
        # Main structure
        self.root_matcher = SM("",
            forwardMatch=True,
            sections=['section_run', "section_frame_sequence", "section_sampling_method",  "section_method"],
            subMatchers=[
                self.cm.header(),
                self.cm.method(),
                self.cm.atoms(),
                self.cm.cell(),
                self.cm.initialization(),
                SM( " DEGREES OF FREEDOM FOR SYSTEM:",
                    sections=["x_cpmd_section_md_initialization"],
                    forwardMatch=True,
                    subMatchers=[
                        SM( " DEGREES OF FREEDOM FOR SYSTEM:"),
                        SM( " RVSCAL| RESCALING IONIC TEMP FROM\s+{0} TO\s+{0}".format(self.regexs.float)),
                        SM( re.escape(" ==                     FORCES INITIALIZATION                  ==")),
                        SM( " EWALD\| SUM IN REAL SPACE OVER\s+{0}\*\s+{0}\*\s+{0} CELLS".format(self.regexs.int)),
                        SM( re.escape(" ==                END OF FORCES INITIALIZATION                ==")),
                        SM( " TIME FOR INITIALIZATION:\s+{} SECONDS".format(self.regexs.float)),
                    ]
                ),
                SM( "       NFI    EKINC   TEMPP           EKS      ECLASSIC          EHAM         DIS    TCPU"),
                SM( re.escape(" *                      AVERAGED QUANTITIES                     *"),
                    sections=["x_cpmd_section_md_averaged_quantities"],
                    subMatchers=[
                        SM( re.escape("                              MEAN VALUE       +/-  RMS DEVIATION")),
                        SM( re.escape("                                     <x>     [<x^2>-<x>^2]**(1/2)")),
                        SM( " ELECTRON KINETIC ENERGY\s+(?P<x_cpmd_electron_kinetic_energy_mean>{0})\s+(?P<x_cpmd_electron_kinetic_energy_std>{0})".format(self.regexs.float)),
                        SM( " IONIC TEMPERATURE\s+(?P<x_cpmd_ionic_temperature_mean>{0})\s+(?P<x_cpmd_ionic_temperature_std>{0})".format(self.regexs.float)),
                        SM( " DENSITY FUNCTIONAL ENERGY\s+(?P<x_cpmd_density_functional_energy_mean>{0})\s+(?P<x_cpmd_density_functional_energy_std>{0})".format(self.regexs.float)),
                        SM( " CLASSICAL ENERGY\s+(?P<x_cpmd_classical_energy_mean>{0})\s+(?P<x_cpmd_classical_energy_std>{0})".format(self.regexs.float)),
                        SM( " CONSERVED ENERGY\s+(?P<x_cpmd_conserved_energy_mean>{0})\s+(?P<x_cpmd_conserved_energy_std>{0})".format(self.regexs.float)),
                        SM( " NOSE ENERGY ELECTRONS\s+(?P<x_cpmd_nose_energy_electrons_mean>{0})\s+(?P<x_cpmd_nose_energy_electrons_std>{0})".format(self.regexs.float)),
                        SM( " NOSE ENERGY IONS\s+(?P<x_cpmd_nose_energy_ions_mean>{0})\s+(?P<x_cpmd_nose_energy_ions_std>{0})              0.000000              0.00000    ".format(self.regexs.float)),
                        SM( " CONSTRAINTS ENERGY\s+(?P<x_cpmd_constraints_energy_mean>{0})\s+(?P<x_cpmd_constraints_energy_std>{0})".format(self.regexs.float)),
                        SM( " RESTRAINTS ENERGY\s+(?P<x_cpmd_restraints_energy_mean>{0})\s+(?P<x_cpmd_restraints_energy_std>{0})".format(self.regexs.float)),
                        SM( " ION DISPLACEMENT\s+(?P<x_cpmd_ion_displacement_mean>{0})\s+(?P<x_cpmd_ion_displacement_std>{0})".format(self.regexs.float)),
                        SM( " CPU TIME\s+(?P<x_cpmd_cpu_time_mean>{0})".format(self.regexs.float)),
                    ]
                ),
                self.cm.footer(),
            ]
        )

    #===========================================================================
    # onClose triggers
    def onClose_section_sampling_method(self, backend, gIndex, section):
        self.sampling_method_gid = gIndex
        backend.addValue("sampling_method", "molecular_dynamics")
        self.cache_service.addValue("ensemble_type")

    def onClose_x_cpmd_section_md_averaged_quantities(self, backend, gIndex, section):
        cons_mean = section.get_latest_value("x_cpmd_conserved_energy_mean")
        cons_std = section.get_latest_value("x_cpmd_conserved_energy_std")
        temp_mean = section.get_latest_value("x_cpmd_ionic_temperature_mean")
        temp_std = section.get_latest_value("x_cpmd_ionic_temperature_std")
        pot_mean = section.get_latest_value("x_cpmd_density_functional_energy_mean")
        pot_std = section.get_latest_value("x_cpmd_density_functional_energy_std")

        self.parse_md()

        if temp_mean is not None and temp_std is not None:
            backend.addArrayValues("frame_sequence_temperature_stats", np.array([temp_mean, temp_std]), unit="K")
        if cons_mean is not None and cons_std is not None:
            backend.addArrayValues("frame_sequence_conserved_quantity_stats", np.array([cons_mean, cons_std]), unit="hartree")
        if pot_mean is not None and pot_std is not None:
            backend.addArrayValues("frame_sequence_potential_energy_stats", np.array([pot_mean, pot_std]), unit="hartree")

    #===========================================================================
    # adHoc
    def parse_md(self):
        """Parses all the md step information.
        """
        # Decide from which file trajectory is read
        n_atoms = self.cache_service["number_of_atoms"]
        trajectory_range = self.cache_service["trajectory_range"]
        trajectory_sample = self.cache_service["trajectory_sample"]
        print_freq = self.cache_service["print_freq"]
        read_trajectory = True
        traj_file = None
        traj_step = 1
        trajec_file_iterator = None
        if self.dcd_filepath is not None:
            traj_file = self.dcd_filepath
        elif self.xyz_filepath is not None:
            traj_file = self.xyz_filepath

        # Initialize the TRAJEC file iterator
        if traj_file is not None:
            try:
                trajec_file_iterator = nomadcore.configurationreading.iread(traj_file)
            except ValueError:
                pass

        # If RANGE is not specified, initialize the TRAJECTORY and FTRAJECTORY
        # iterators if files available
        ftrajectory_file_iterator = None
        trajectory_file_iterator = None
        if not trajectory_range:
            if self.ftrajectory_filepath is not None:
                ftrajectory_file_iterator = nomadcore.csvparsing.iread(self.ftrajectory_filepath, columns=range(10), n_conf=n_atoms)
            if self.trajectory_filepath is not None:
                trajectory_file_iterator = nomadcore.csvparsing.iread(self.trajectory_filepath, columns=range(7), n_conf=n_atoms)

        # Initialize the ENERGIES file iterator
        energies_iterator = None
        if self.energies_filepath is not None:
            energies_iterator = nomadcore.csvparsing.iread(self.energies_filepath, columns=range(8))

        # Start reading the frames
        i_frame = 0
        n_frames = self.cache_service["n_steps"]
        time_step = self.cache_service["time_step_ions"]
        self.backend.addArrayValues("frame_sequence_time", np.array([(x+1)*time_step for x in range(n_frames)]))
        self.backend.addValue("number_of_frames_in_sequence", n_frames)

        temperatures = []
        potential_energies = []
        kinetic_energies = []
        conserved_quantities = []

        for i_frame in range(n_frames):

            # Open sections
            scc_id = self.backend.openSection("section_single_configuration_calculation")
            sys_id = self.backend.openSection("section_system")

            # System
            self.cache_service.addArrayValues("atom_labels")
            self.cache_service.addArrayValues("simulation_cell", unit="bohr")
            self.cache_service.addValue("number_of_atoms")

            if print_freq is not None:
                if i_frame % print_freq == 0:

                    # TRAJEC file
                    if trajec_file_iterator is not None:
                        try:
                            pos = next(trajec_file_iterator)
                        except StopIteration:
                            LOGGER.error("Could not get the next geometries from a TRAJEC file.")
                        else:
                            self.backend.addArrayValues("atom_positions", pos, unit="angstrom")

                    # FTRAJECTORY file
                    if ftrajectory_file_iterator is not None:
                        try:
                            values = next(ftrajectory_file_iterator)
                        except StopIteration:
                            LOGGER.error("Could not get the next configuration from a FTRAJECTORY file.")
                        else:
                            velocities = values[:, 4:7]
                            self.backend.addArrayValues("atom_velocities", velocities, unit="bohr/(hbar/hartree)")

                            forces = values[:, 7:10]
                            self.backend.addArrayValues("atom_forces", forces, unit="hartree / bohr")

                            if trajec_file_iterator is None:
                                pos = values[:, 1:4]
                                self.backend.addArrayValues("atom_positions", pos, unit="bohr")

                    # TRAJECTORY file
                    if ftrajectory_file_iterator is None:
                        if trajectory_file_iterator is not None:
                            try:
                                values = next(trajectory_file_iterator)
                            except StopIteration:
                                LOGGER.error("Could not get the next configuration from a TRAJECTORY file.")
                            else:
                                velocities = values[:, 4:]
                                self.backend.addArrayValues("atom_velocities", velocities, unit="bohr/(hbar/hartree)")

                                if trajec_file_iterator is None:
                                    pos = values[:, 1:4]
                                    self.backend.addArrayValues("atom_positions", pos, unit="bohr")

            # Energies file
            if energies_iterator is not None:
                try:
                    values = next(energies_iterator)
                except StopIteration:
                    LOGGER.error("Could not get the next configuration from an ENERGIES file.")
                else:
                    potential_energy = values[3]
                    conserved_quantity = values[5]
                    ion_total_energy = values[4]
                    kinetic_energy = ion_total_energy - potential_energy
                    temperature = values[2]
                    # electronic_kinetic_energy = values[1]
                    # msd = values[6]
                    # i_step = values[0]
                    tcpu = values[7]
                    conserved_quantities.append(conserved_quantity)
                    potential_energies.append(potential_energy)
                    kinetic_energies.append(kinetic_energy)
                    temperatures.append(temperature)
                    self.backend.addRealValue("energy_total", potential_energy, unit="hartree")
                    self.backend.addValue("time_calculation", tcpu)

            # Close sections
            self.backend.closeSection("section_single_configuration_calculation", scc_id)
            self.backend.closeSection("section_system", sys_id)

        # Push the summaries
        if potential_energies:
            potential_energies = np.array(potential_energies)
            self.backend.addArrayValues("frame_sequence_potential_energy", potential_energies, unit="hartree")
        if kinetic_energies:
            kinetic_energies = np.array(kinetic_energies)
            self.backend.addArrayValues("frame_sequence_kinetic_energy", kinetic_energies, unit="hartree")

            # Push the statistics. CPMD prints some statistics at the end, but the
            # mean and std of kinetic energy are missing
            kin_mean = kinetic_energies.mean()
            kin_temp = (kinetic_energies - kin_mean)
            kin_std = np.sqrt(np.dot(kin_temp, kin_temp)/kinetic_energies.size)
            kin_temp = None
            self.backend.addArrayValues("frame_sequence_kinetic_energy_stats", np.array([kin_mean, kin_std]), unit="hartree")

        if conserved_quantities:
            conserved_quantities = np.array(conserved_quantities)
            self.backend.addArrayValues("frame_sequence_conserved_quantity", conserved_quantities, unit="hartree")
        if temperatures:
            temperatures = np.array(temperatures)
            self.backend.addArrayValues("frame_sequence_temperature", temperatures, unit="K")
